import time
from web.server import HttpServer, request
from web.page import *
import os


class TestClass:

    def test(self):
        return request.url

    def index(self):
        return Body(H1("Welcome To Me"))

    def hello(self, id, word="me"):
        return "Hello " + str(id) + ", word = " + str(word)

    def dict(self, name):
        print('in dict', request.args)
        return {'name': name, 'result': 0}

    def list(self):
        return [1, 88, 34]

    def number(self):
        return 99

    def exception(self):
        raise ValueError("Val invalid")


def serve():
    obj = TestClass()
    server = HttpServer(debug=True)
    # path = os.path.join(os.path.dirname(__file__), "www")
    # print(path)
    # server.static_folder("www")
    server.add_url('/', obj).run(threaded=True)
    print("start")
    time.sleep(300)
    print("stopping")
    server.stop()
    print("finished")


serve()
# Html(Body(Div("This is a page"))).save_html("www/test1.html")





#
# Simplify website creation
#
import re
from pscript import py2js
import inspect
import random
import json
from future.backports import html

# a fack apps in python
apps = None


# function for javascript
def main():
    Apps(apps).run()


# configurations
class Config:
    """ configurations """
    jquery_url = "http://libs.baidu.com/jquery/2.0.0/jquery.min.js"


class PageUtil:
    @staticmethod
    def html(obj, indent=0):
        """ convert object to html """
        if hasattr(obj, 'html'):
            return obj.html(indent)
        else:
            return str(obj) if obj is not None else ''

    @staticmethod
    def cameled(prop):
        """ convert property name to camel cased """
        return re.sub(r'-([a-z])', lambda m: m.group(1).upper(), prop)

    @staticmethod
    def dashed(prop):
        """ convert property name to dashed string """
        return re.sub(r'[A-Z]', lambda m: '-' + m.group().lower(), prop)

    @staticmethod
    def trim_var_quote(var_name):
        """ trim variable quote marks """
        if var_name.startswith('{') and var_name.endswith('}'):
            return var_name[1:len(var_name)-1]
        return var_name

    @staticmethod
    def random_str(prefix='', length=8):
        """ return a random string """
        ascii_lowercase = 'abcdefghijklmnopqrstuvwxyz'
        return prefix + ''.join(random.sample(ascii_lowercase, length))


class Variable:
    def __init__(self, name, value=None):
        self.name = name
        self.value = value

    def html(self, indent=0):
        val = self.value if self.value is not None else ''
        return '%s' % val

    def __repr__(self):
        return "<Variable %s>" % repr(self.name)


class Expression:

    @staticmethod
    def parse(expr_str=None, element=None):
        ex = Expression(expr_str, element)
        if len(ex.items) == 0:
            return ''
        elif len(ex.items) == 1 and isinstance(ex.items[0], str):
            return ex.items[0]
        else:
            return ex

    def __init__(self, expr=None, element=None):
        self.element = element
        self.vars = {}
        self.items = []
        if expr:
            self.load(expr)

    def load(self, expr):
        if not isinstance(expr, str):
            self.items.append(expr)
            return

        var_started = False
        varname = ""
        text = ''
        last_c = ''
        for index, c in enumerate(expr):
            if c == '{' and last_c == '{':
                text = text[:len(text) - 1]
                if text:
                    self.items.append(text)
                text = ''
                var_started = True
                varname = ''
                continue
            elif c == '}' and last_c == '}':
                varname = varname[:len(varname) - 1]
                self.items.append(Variable(varname))
                var_started = False
                varname = ''
                continue

            if var_started:
                varname += c
            else:
                text += c
            last_c = c

        if text:
            self.items.append(text)

        return self
        # if len(self.items) <= 0:
        #     return ''
        # elif len(self.items) == 1 and isinstance(self.items[0], str):
        #     return self.items[0]
        # else:
        #     return self

    def get_vars(self):
        var_list = []
        for item in self.items:
            if isinstance(item, Variable):
                var_list.append(item)
        return var_list

    def html(self, indent=0):
        ret = ''
        for item in self.items:
            ret += PageUtil.html(item)
        return ret

    def js_string(self):
        ret = ''
        for index, item in enumerate(self.items):
            if index > 0:
                ret += '`~`'

            if isinstance(item, Variable):
                ret += '{{' + item.name + '}}'
            else:
                ret += str(item)
        return html.escape(ret)

    def __repr__(self):
        ret = '['
        for index, item in enumerate(self.items):
            if index > 0:
                ret += ', '
            if isinstance(item, Variable):
                ret += '{{' + item.name + '}}' if item.name is not None else ''
            else:
                ret += str(item)
        ret += ']'
        return ret


class Children(list):
    def __init__(self, parent_element, *children):
        super().__init__()
        self.element = parent_element
        if children:
            self.add(children)

    def find(self, element):
        """ find a child element, return index """
        for index, child in enumerate(self):
            if element == child:
                return index

    def add(self, child, raw=False):
        """
        add a child element

        :param child: child element or a string
        :param raw:   (optional) whether the child is raw text
        :return: self
        """
        if not child:
            return self

        if isinstance(child, tuple):
            for c in child:
                self.add(c)
            return self

        if not isinstance(child, Element):
            child = Text(child) if not raw else child
            self.append(child)
            if isinstance(child, Element):
                child['@parent'] = self.element
        elif isinstance(child, Element):
            self.append(child)
            child['@parent'] = self.element

        return self

    def delete(self, element):
        """
        delete a child element

        :param element: element object or index int
        :return: self
        """
        if isinstance(element, int):
            if 0 <= element < len(self):
                element = self[element]
            else:
                return self

        for child in self:
            if element == child:
                self.remove(element)
                if isinstance(element, Element):
                    element['@parent'] = None
        return self

    def html(self, indent=0):
        """ convert to html """
        crlf = '\r\n' if indent > 0 else ''

        ret = ''
        for item in self:
            if isinstance(item, Element):
                val = item.html(indent) + crlf
                ret += val if val is not None else ''
            else:
                ret += str(item) if item is not None else ''
        return ret


class Element(dict):
    """ A Element is an element of HTML page """

    def __init__(self, *children, **properties):
        """
        init, add children elements, set options.

        :param children: children elements
        :param properties:  properties of the element.
        """
        super().__init__()
        tag = self._get_cls_attr('@tag', True)
        self['@tag'] = tag if tag else type(self).__name__.lower()
        self['@children'] = Children(self)
        self['@parent'] = None

        # add content from template
        template = self._get_cls_attr('template')
        if template:
            self.add_child(template)

        # add props
        props = self._get_cls_attr('props')
        if isinstance(props, dict):
            for k in props:
                if k == '@text':
                    self.add_child(props[k], raw=True)
                else:
                    self.add_prop(k, props[k])

        # add child from children
        for child in children:
            self.add_child(child)

        # add properties form kwargs
        for k in properties:
            if k == 'text':
                self.add_child(properties[k], raw=True)
            else:
                self.add_prop(k, properties[k])

    @property
    def parent(self):
        return self['@parent']

    @property
    def children(self):
        return self['@children']

    @property
    def tag(self):
        """ the tag of the element """
        return self['@tag']

    def _get_cls_attr(self, name, find_parent=True):
        """ get class static attribute of specified name """
        def get_parent_cls(cls):
            if len(cls.__bases__) > 0:
                return cls.__bases__[0]

        def find_cls(cls):
            """ find attribute in specified class """
            val = cls.__dict__.get(name, None)
            if val:
                return val

            # if find in parent is needed
            if find_parent and cls != Element:
                parent_cls = get_parent_cls(cls)
                # if parent class is Element and find '@tag'
                if parent_cls == Element and name == '@tag':
                    return cls.__name__.lower()

                # find in parent class
                if parent_cls == object:
                    return None
                elif parent_cls is not None:
                    return find_cls(parent_cls)

        return find_cls(type(self))

    @property
    def _root_parent(self):
        """ get the root parent """
        p = self
        while p.parent:
            p = p.parent
        return p

    def _html_props(self):
        """ convert properties to html """
        ret = ''
        for item in self:
            if item.startswith('@'):
                continue
            val = self[item]
            if item == 'expr_props' and isinstance(val, list):
                val = '`~`'.join(val)
            item_name = 'class' if item == 'clazz' else item
            ret += ' '
            ret += PageUtil.dashed(item_name) + '="' + html.escape(PageUtil.html(val)) + '"'
        return ret

    def add_prop(self, prop, value):
        """
        add a property

        :param prop:  property name
        :param value: property value
        :return: self
        """
        expr = Expression.parse(value)
        self[prop] = expr
        if isinstance(expr, Expression):
            var_list = expr.get_vars()
            if var_list:
                expr_props = self.get('expr_props', [])
                if prop not in expr_props:
                    expr_props.append(prop)
                self['expr_' + prop] = expr.js_string()
                self['expr_props'] = expr_props
        return self

    def add_child(self, child, raw=False):
        """
        add a child element

        :param child: child element or a string
        :param raw:   (optional) whether the child is raw text
        :return: self
        """
        self.children.add(child, raw=raw)
        return self

    def add_script(self, *codes, **kwargs):
        """
        add script

        :param codes:  list of code, code could be str or Script object
        :param kwargs: input arguments for the script
        :return: self
        """
        def js_code(name, value):
            """ convert to javascript statement 'var name=value' """
            if isinstance(value, list) or isinstance(value, dict):
                return 'var %s=%s;\n' % (name, json.dumps(value))
            elif isinstance(value, str):
                return 'var %s=%s;\n' % (name, repr(value))
            elif isinstance(value, int) or isinstance(value, float):
                return 'var %s=%s;\n' % (name, value)
            elif value is None:
                return 'var %s=null;' % name
            else:
                return 'var %s=%s;\n' % (name, str(value))

        # code list
        code_list = []

        # if not init in root element
        root = self._root_parent
        if not root.get('@init'):
            # add jquery script
            self.add_child(Script(src=Config.jquery_url))
            # add class and function for javascript
            code_list = code_list + [App, Apps, getElement]
            root['@init'] = True

        # add arguments
        for index, key in enumerate(kwargs):
            if index == 0:
                code_list.append("//arguments\n")
            code_list.append((js_code(key, kwargs[key])))

        # add codes
        code_list = code_list + list(codes)
        # create Script(object)
        self.add_child(Script(*code_list))

    def html(self, indent=0):
        """ convert to html """
        prefix = ''
        crlf = ''

        if indent is True:
            indent = 1

        if indent > 0:
            crlf = '\n'
            prefix = ' ' * (indent - 1) * 2
            indent += 1

        if self.get('@tag', None):
            ret = ''
            ret += prefix + '<' + self.tag + self._html_props() + '>' + crlf
            if len(self.children) > 0:
                ret += self.children.html(indent)
            ret += prefix + '</' + self.tag + '>'
            return ret
        else:
            return self.children.html(indent)

    def get_vars(self):
        """return a list of Variable declared within the element """
        var_list = []
        # set data of props
        for item in self:
            if item in ['tag', 'parent', 'children', '_data']:
                continue
            else:
                val = self[item]
                if isinstance(val, Expression):
                    var_list += val.get_vars()

        # set data of children
        for child in self.children:
            if isinstance(child, Element):
                var_list += child.get_vars()

        return var_list

    def __repr__(self):
        return type(self).__name__

    def __getattr__(self, item):
        if item in self:
            return self[item]

        for child in self.children:
            if isinstance(child, Element):
                if child.get('id') == item:
                    return child

        for child in self.children:
            if isinstance(child, Element):
                if child.get('@tag') == item:
                    return child

        if item == 'clazz':
            item = 'class'
        for child in self.children:
            if isinstance(child, Element):
                cls = child.get('class', '')
                if (cls + ' ').find(item+' ') >= 0:
                    return child

        return None

    def walk(self, func):
        """
        walk through the children element, for each element call function

        :param func: the function which should accept one parameter
        :return: self
        """
        func(self)
        for item in self.children:
            if isinstance(item, Element):
                item.walk(func)
        return self

    def save_html(self, filename, apps=None, encoding='utf-8'):
        """
        save element to html file

        :param filename:  html filename
        :param apps:  (optional) a diction which defines app_ids
        :param encoding:  (optional) encoding of the file
        :return: self
        """
        if apps:
            self.add_script(main, apps=apps)

        with open(filename, 'w', encoding=encoding) as f:
            f.write(self.html(True))


class NoContentElement(Element):
    def html(self, indent=0):
        return '<%s%s/>' % (self.tag, self._html_props())


class Text(Element):
    tag = 'span'

    def __init__(self, text):
        super().__init__()
        expr = Expression.parse(text)
        self['@text'] = expr
        self['@has_var'] = False

        if isinstance(expr, Expression):
            self['expr_text'] = expr.js_string()
            self['@has_var'] = True

    def html(self, indent=0):
        if self['@has_var']:
            return super().html(indent)
        else:
            return PageUtil.html(self['@text'], indent)


class Comment(Element):
    def __init__(self, text):
        super().__init__()
        self['@text'] = text

    def html(self, indent=0):
        """ convert to html """
        if indent is True:
            indent = 1
        prefix = '  ' * (indent - 1) if indent > 1 else ''
        return prefix + '<!-- ' + html.escape(self['@text']) + ' -->'


class Script(Element):
    def __init__(self, *children, **properties):
        super().__init__(**properties)

        if len(children) == 1 and isinstance(children[0], str):
            self['src'] = children[0]
            return

        last_func = ''
        stdlib = True
        for item in children:
            if isinstance(item, str):
                self.add_child('\n' + item, raw=True)
            elif inspect.isfunction(item):
                self.add_child('\n' + py2js(item), raw=True)
                last_func = item.__name__
            elif inspect.isclass(item):
                self.add_child('\n' + py2js(item, inline_stdlib=stdlib), raw=True)
                stdlib = False

        if last_func:
            self.add_child('\njQuery.ready(' + last_func + '());\n')


class A(Element):pass
class B(Element):pass
class Abbr(Element): pass
class Acronym(Element): pass
class Address(Element): pass
class Applet(Element): pass
class Area(Element): pass
class Article(Element): pass
class Aside(Element): pass
class Audio(Element): pass
class Base(Element): pass
class Basefont(Element): pass
class Bdi(Element): pass
class Bdo(Element): pass
class Big(Element): pass
class Blockquote(Element): pass
class Body(Element): pass


class Br(NoContentElement):
    tag = 'br'
    pass

class Button(Element): pass
class Canvas(Element): pass
class Caption(Element): pass
class Center(Element): pass
class Code(Element): pass
class Col(Element): pass
class Command(Element): pass
class Dd(Element): pass
class Del(Element): pass
class Details(Element): pass
class Dir(Element): pass
class Div(Element): pass
class Dialog(Element): pass
class Dl(Element): pass
class Dt(Element): pass
class Em(Element): pass
class Embed(Element): pass
class Fieldset(Element): pass
class Figcaption(Element): pass
class Figure(Element): pass
class Font(Element): pass
class Footer(Element): pass
class Form(Element): pass
class Frame(Element): pass
class Frameset(Element): pass
class H1(Element): pass
class H2(Element): pass
class H3(Element): pass
class H4(Element): pass
class H5(Element): pass
class H6(Element): pass
class Head(Element): pass
class Header(Element): pass


class Hr(NoContentElement):
    tag = 'hr'


class Html(Element): pass
class Iframe(Element): pass
class Img(Element): pass


class Input(NoContentElement):
    tag = 'input'


class Label(Element): pass
class Legend(Element): pass
class Li(Element): pass
class Link(Element): pass
class Mark(Element): pass
class Menu(Element): pass
class Menuitem(Element): pass


class Meta(NoContentElement):
    tag = 'meta'


class Nav(Element): pass
class Noframes(Element): pass
class Noscript(Element): pass
class Object(Element): pass
class Ol(Element): pass
class Optgroup(Element): pass
class Option(Element): pass
class Output(Element): pass
class Param(Element): pass
class Pre(Element): pass
class Progress(Element): pass
class Rp(Element): pass
class Rt(Element): pass
class Ruby(Element): pass
class Samp(Element): pass
class Section(Element): pass
class Select(Element): pass
class Small(Element): pass
class Source(Element): pass
class Span(Element): pass
class Strike(Element): pass
class Strong(Element): pass
class Style(Element): pass
class Sub(Element): pass
class Summary(Element): pass
class Sup(Element): pass
class Svg(Element): pass
class Table(Element): pass
class Tbody(Element): pass
class Td(Element): pass
class Template(Element): pass
class Textarea(Element): pass
class Tfoot(Element): pass
class Th(Element): pass
class Thead(Element): pass
class Time(Element): pass
class Title(Element): pass
class Tr(Element): pass
class Track(Element): pass
class Tt(Element): pass
class U(Element): pass
class Ul(Element): pass
class Var(Element): pass
class Video(Element): pass
class Wbr(Element): pass
class Xmp(Element): pass


# a fake jQuery under python
class jQuery:
    def __init__(self, obj):
        self.obj = obj

    def attr(self, name, value=None):
        if value is None:
            return self.obj[name]
        else:
            self.obj[name] = value

    def html(self, text=None):
        if text is not None:
            children = self.obj.children
            if len(children) == 1 and isinstance(children[0], Text):
                children[0].add_child(text)
            else:
                children.clear()
                children.add_child(text)
        else:
            return self.obj.children.html()


# function for javascript
def getElement(selector):
    """ this function need to call under javascript """
    return jQuery(selector)


# class for javascript
class App:
    """
    An App is an element with app_id property in the html.
    Data is bound to the children element.
    """
    def __init__(self, name):
        self.name = name
        self.elem = getElement('[app_id=%s]' % name)
        if not self.elem:
            raise Exception("App %s not found" % repr(name))

    def html_decode(self, value):
        """ decode html entities """
        if value:
            return jQuery('<div />').html(value).text()
        else:
            return ''

    def html_encode(self, value):
        """ encode html entities """
        if value:
            return jQuery('<div />').text(value).html()
        else:
            return ''

    def _is_int(self, s):
        """ whether the string is an integer """
        for c in s:
            if '0' <= c <= '9':
                pass
            else:
                return False
        return True

    def _dict_value(self, data, key, default_val=''):
        """ get value from dictionary by key. key could be xxx.xxx.xxx """
        keys = key.split('.')
        d = data
        for k in keys:
            if isinstance(d, dict) and k in d:
                d = d[k]
            elif isinstance(d, list) and self._is_int(k) and len(d) > int(k):
                d = d[int(k)]
            else:
                return default_val
        return d

    def _expression_value(self, expr_text, data):
        """ get expression value from data """
        text = ''
        if expr_text:
            expr_text = self.html_decode(expr_text)
            items = expr_text.split('`~`')
            for item in items:
                if item.startswith('{{') and item.endswith('}}'):
                    item = item[2:len(item) - 2]
                    item = self._dict_value(data, item)
                text += item
        return text

    def _set_element_text(self, e, data):
        """ update variable value in the element text from the data """
        expr_text = e.attr('expr_text')
        text = self._expression_value(expr_text, data)
        tag = e.context.nodeName.lower()
        if tag in ['pre', 'code', 'input', 'textarea']:
            e.text(text)
        else:
            text = self.html_encode(text)
            text = text.replace('\n', '<br>')
            e.html(text)

    def set_data(self, data):
        """ update variable value from the data """
        # set content text
        elems = self.elem.find('[expr_text]')
        if self.elem.attr('expr_text'):
            elems.push(self.elem)
        for i in range(elems.length):
            elem = getElement(elems[i])
            self._set_element_text(elem, data)

        # set props
        elems = self.elem.find('[expr_props]')
        if self.elem.attr('expr_props'):
            elems.push(self.elem)
        for i in range(elems.length):
            elem = getElement(elems[i])
            props = elem.attr('expr_props')
            props = props.split('`~`')
            for prop in props:
                expr_text = elem.attr('expr_' + prop)
                text = self._expression_value(expr_text, data)
                elem.attr(prop, text)

    def set_http(self, url):
        """ update variable value from the data which is get from the url """
        def on_success(result):
            self.set_data(result)

        jQuery.ajax({'url': url, 'dataType': 'JSON', 'success': on_success})


# class for javascript
class Apps:
    """ Apps"""
    def __init__(self, apps):
        """
        init the apps definition

        :param apps:  A dict of App
        """
        self.input = apps
        self.apps = {}

    def run(self):
        """ run the apps """
        def on_success(result):
            self.apps = result
            self.update()

        if isinstance(self.input, dict):
            self.apps = self.input
            self.update()
        elif isinstance(self.input, str):
            # get apps definition from the url
            jQuery.ajax({'url': self.input, 'dataType': 'JSON', 'success': on_success})

    def update(self):
        """ update all app """
        if isinstance(self.apps, dict):
            for app_id in self.apps:
                data = self.apps[app_id]
                if isinstance(data, dict):
                    App(app_id).set_data(data)
                elif isinstance(data, str):
                    App(app_id).set_http(data)




import net


url = 'https://www.w3school.com.cn/tags/index.asp'
tables = net.Spider(url).find_tables(text_only=True)
table = tables[0]
for row in table:
    tag = row[0].text
    if tag.startswith('<') and tag.endswith('>'):
        tag = tag[1:len(tag)-1]
        if len(tag) > 1:
            c = tag[0].upper()
            tag = c + tag[1:]
            if not tag.startswith('!'):
                print("class %s(Element): pass" % tag)



import unittest
from web.page import *


class TestUtil(unittest.TestCase):
    def test_dashed(self):
        s = "text-color"
        dashed = PageUtil.dashed(s)
        self.assertEqual(dashed, 'text-color')
        cameled = PageUtil.cameled(dashed)
        self.assertEqual(cameled, 'textColor')

        s = "TextColor"
        dashed = PageUtil.dashed(s)
        self.assertEqual(dashed, '-text-color')
        cameled = PageUtil.cameled(dashed)
        self.assertEqual(cameled, 'TextColor')

        s = "color"
        dashed = PageUtil.dashed(s)
        self.assertEqual(dashed, 'color')
        cameled = PageUtil.cameled(dashed)
        self.assertEqual(cameled, 'color')

    def test_html(self):
        pass


class TestExpression(unittest.TestCase):
    def test_parse(self):
        ret = Expression("hello")
        self.assertEqual(len(ret.items), 1)
        self.assertEqual(ret.items[0], "hello")

        ret = Expression("hello, {{name}}")
        self.assertEqual(len(ret.items), 2)
        self.assertEqual(ret.items[0], "hello, ")
        self.assertEqual(type(ret.items[1]), Variable)


class TestElement(unittest.TestCase):
    def test1(self):
        self.assertEqual(Element().html(), '<element></element>')
        self.assertEqual(Html().html(), '<html></html>')
        self.assertEqual(Div("text").html(), '<div>text</div>')
        self.assertEqual(Div(id=3).html(), '<div id="3"></div>')
        self.assertEqual(Br().html(), '<br/>')
        self.assertEqual(Meta(charset="utf-8").html(), '<meta charset="utf-8"/>')


if __name__ in "__main__":
    unittest.main()



from pscript import py2js
from web.page import *


window = None
document = None
jQuery = None


class TestA:
    def __init__(self, name, age):
        self.name = name
        self.age = age

    def hello(self):
        print("hello", self.name)


class TestB:
    def m2(self):
        print("me")


def main():
    # TestA("Peter", 18).hello()
    Object("#title").html("Hello,你好")


class Page(Element):
    tag = 'html'
    template = Head(), Body()


import random


page = Html(
    Head(),
    Body(
        Div(id="hello"),
        Div(clazz='abb')
    )
)

# print(PageUtil.random_str())

page = Body(Div("ttt"))
# page.body.add("Pure")
# page.body.\
# page.add(Div("{{content}}", clazz="{{clz}}"))

# page.body.add(Div("{{message}}", id="title"))
# vars = page.body.get_vars()
# print(vars)
print(page.html())
exit()

# page.bind_data({'message': 'Hello, world'})
# page.body.set_data({'message': 'Hello, world'})
# page.add_script(TestA, TestB, main)
# page.add_script(main)
print(page.html())
save_html(page, "test2.html")


# page.head.append(script)

# import inspect
#
# save_html(page, "test2.html")
# # print(inspect.isclass(TestA))
#
# print(inspect.isfunction(start_func))
# item = start_func
# print(item.__name__)
# print(inspect.ismethod(start_func))
# print(inspect.ismethod(TestA.hello))

# print(py2js(TestA))



from web.page import *


class Button(Div):
    props = {"style": "color:red;", "@text": "Button"}


e = Div("Hello \n{{message}}", app_id="app1", style="color:{{color}};", id="message")
page = Html(Head(), Body(e, Button()))
# page.body.children.delete(page.body.div)
# page.body.children.delete(page.body.div)
page.body.add_child(Comment("this is <a> comment"))

apps = {"app1": {"message": "World", "color": "red"}}
apps = {"app1": "/jojo-web/data.json"}
apps = "/jojo-web/apps.json"
page.save_html("test2.html", apps=apps)



from pscript import py2js



class TestA:
    def __init__(self, name, age):
        self.name = name
        self.age = age

    def hello(self):
        print("hello", self.name)


class TestB(TestA):
    def hello2(self):
        print("hello2 from TestB", self.name, self.age)


def foo(a, b=2):
   print(a - b)



print(py2js(TestB))




# from collections import OrderedDict
#
#
# class PageUtil:
#     @staticmethod
#     def is_list_like(obj):
#         if isinstance(obj, list) or isinstance(obj, tuple):
#             return True
#         return False
#
#     @staticmethod
#     def to_string(obj):
#         s = str(obj)
#         if s.find('"') >= 0:
#             if s.find("'") >= 0:
#                 s = s.replace('"', '&quot;')
#                 return '"' + s + '"'
#             else:
#                 return "'" + s + "'"
#         else:
#             return '"' + s + '"'
#
#
# class Clazz:
#     @staticmethod
#     def is_child(child_cls, parent_cls):
#         """ whether child_cls is the child class of parent_cls"""
#         if child_cls == parent_cls:
#             return True
#
#         bases = child_cls.__bases__
#         for base_cls in bases:
#             if Clazz.is_child(parent_cls, base_cls):
#                 return True
#
#     @staticmethod
#     def find_attr(cls, attr):
#         """ find static attribute value in cls and its parent class """
#         if attr in cls.__dict__:
#             # the cls contains attr
#             return cls.__dict__[attr]
#         else:
#             # find attr in parent class
#             bases = cls.__bases__
#             for base_cls in bases:
#                 if Clazz.is_child(base_cls, Element):
#                     ret = Clazz.find_attr(base_cls, attr)
#                     if ret:
#                         return ret
#
#
# class Props(dict):
#     def __init__(self, element):
#         self.element = element
#         super().__init__()
#
#     def add(self, props):
#         if props is None:
#             return self
#
#         if isinstance(props, dict):
#             self.element._set_options(props)
#             return self
#
#         if not isinstance(props, list) and not isinstance(props, tuple):
#             props = [props]
#
#         if isinstance(props, list) or isinstance(props, tuple):
#             for p in props:
#                 self[p] = ''
#
#         return self
#
#     def gen(self, indent=0):
#         ret = ' ' if len(self) > 0 else ''
#         for index, p in enumerate(self):
#             if index > 0:
#                 ret += ' '
#             ret += str(p) + '=' + PageUtil.to_string(self[p])
#         return ret
#
#     def __repr__(self):
#         return repr(self.keys())
#
#
# class Classes(list):
#     def __init__(self, element):
#         self.element = element
#         super().__init__()
#
#
# class Element:
#     tag = 'div'
#
#     def __init__(self, *args, **kwargs):
#         self._tag = Clazz.find_attr(type(self), 'tag') or 'div'
#         self._no_content = False
#         self._classes = Classes(self)
#         self._props = Props(self)
#         self._children = Children(self)
#         self._parent = None
#
#         if len(args) == 1:
#             if isinstance(args[0], dict):
#                 kwargs = args[0]
#                 args = ()
#
#         self._set_no_content()
#         self._create_template()
#         self._create_props()
#         self._children.add(args)
#         self._set_options(kwargs)
#
#     def _set_no_content(self):
#         if 'no_content' in self.__class__.__dict__:
#             self._no_content = True
#
#     def _create_template(self):
#         if 'template' in self.__class__.__dict__:
#             template = self.__class__.__dict__['template']
#             self.children.add(template)
#         # template = Clazz.find_attr(type(self), 'template')
#         # self.children.add(template)
#
#     def _create_props(self):
#         p = Clazz.find_attr(type(self), 'props')
#         if p:
#             if isinstance(p, dict):
#                 self._set_options(p)
#             else:
#                 self.prop.add(p)
#
#     def _set_options(self, options):
#         for key in options:
#             if key in ['cls', 'class']:
#                 key = 'class'
#                 self._props[key] = options[key]
#             else:
#                 self._props[key] = options[key]
#         return self
#
#     @property
#     def children(self):
#         return self._children
#
#     @property
#     def prop(self):
#         return self._props
#
#     def _gen_props(self, indent=0):
#         return self._props.gen(indent)
#
#     def _gen_content(self, indent=0):
#         ret = ''
#         indent = 0
#         for child in self._children:
#             ret += child.gen(indent)
#         return ret
#
#     def gen(self, indent=0):
#         if self._no_content:
#             s = self._props.gen(indent)
#             if s:
#                 return '<%s%s />' % (self._tag, s)
#             else:
#                 return '<%s>' % self._tag
#         else:
#             return '<%s%s>%s</%s>' % (self._tag, self._props.gen(indent),
#                                       self._gen_content(indent), self._tag)
#
#     def __str__(self):
#         return self.gen()
#
#
# class Text(Element):
#     def __init__(self, *args, **kwargs):
#         self.args = args
#         super().__init__(**kwargs)
#
#     def gen(self, indent=0):
#         ret = ''
#         for arg in self.args:
#             if arg is not None:
#                 ret += str(arg)
#         return ret
#
#
# class TITLE(Element):
#     tag = 'title'
#
#
# class META(Element):
#     tag = 'meta'
#     no_content = True
#
#
# class CharSet(Element):
#     tag = 'meta'
#     no_content = True
#     props = {'http-equiv': "Content-Type", 'content': "text/html; charset=UTF-8"}
#
#     def __init__(self, encoding=None):
#         super().__init__()
#         if encoding:
#             self._props['content'] = "text/html; charset=%s" % encoding
#
#
# class BR(Element):
#     tag = 'br'
#     no_content = True
#
#
# class HEAD(Element):
#     tag = 'head'
#     template = TITLE()
#
#
# class BODY(Element):
#     tag = 'body'
#
#
# class DIV(Element):
#     tag = 'div'
#
#
# class H1(Element):
#     tag = 'h1'
#
#
# class H2(Element):
#     tag = 'h2'
#
#
# class H3(Element):
#     tag = 'h3'
#
#
# class H4(Element):
#     tag = 'h4'
#
#
# class H5(Element):
#     tag = 'h6'
#
#
# class SPAN(Element):
#     tag = 'span'
#
#
# class A(Element):
#     tag = 'a'
#
#
# class IMG(Element):
#     tag = 'img'
#
#
# class Component(Element):
#     template = None
#
#
# class Page(Component):
#     tag = 'html'
#     template = HEAD(), BODY()
#
#
# class Child(H1):
#     props = ['message']
#
#
# if __name__ == "__main__":
#     # e = H1(A(IMG(src="http://abc.com/1.png"), href='http://www.baidu.com'))
#     # print(e)
#
#     # a = META({'http-equiv': "Content-Type", 'content': "text/html; charset=UTF-8"})
#     # a = CharSet('GB2312')
#     # print(a)
#
#     p = Page(BR(), "Hello, World")
#     # p.head.title = "hello"
#     print(p)
#
#     # r = Child("自定义组件!")
#     # r.prop['message'] = "Hello"
#     # print(r)
#
#
#



import flask
from flask import Flask
from werkzeug.serving import make_server

import threading
import logging
import inspect
from inspect import Parameter
from collections import OrderedDict
import os
import re


# request from flask
request = flask.request

# render_template from flask
render_template = flask.render_template


class HttpServer:
    """
    Http server.

    Example:

    HttpServer(port=8080).run(threaded=True)

    """
    WelcomeHtml = '<html><body><h1>Welcome</h1></body></html>'

    FileExts = ['.txt', '.htm', '.html', '.jpg', '.jpeg', '.css', '.gif', '.png',
                '.wav', '.mp3', '.mp4', '.mpg', '.mpeg', '.m3u', '.flv',
                '.json', '.js', '.doc', '.docx', '.ppt', '.pptx', '.xls', '.xlsx']

    def __init__(self, port=80, ip='0.0.0.0', debug=False, app=None):
        """
        Http server

        :param port:  server post
        :param ip:    ip address, '0.0.0.0' or '127.0.0.1'
        :param debug: (optional) whether in debug mode
        :param app:   (optional) flask app object
        """
        self.ip = ip if isinstance(ip, str) else '0.0.0.0'
        self.port = port
        self.app: flask.Flask = app  # flask app
        self.debug = debug
        self._server = None
        self._has_root = False
        self._static_folder = None
        self._static_url_path = None

    def _confirm_app(self):
        """ confirm self.app is initialized """
        if self.app is None:
            self.app = Flask(__name__)

    def _get_func_params(self, func):
        """ get func parameters ans its values, return a dict """
        defaults = OrderedDict()
        if inspect.ismethod(func) or inspect.isfunction(func):
            params = inspect.signature(func).parameters
            for i in params:
                if params[i].default == Parameter.empty:
                    defaults[params[i].name] = None
                else:
                    defaults[params[i].name] = params[i].default
        return defaults

    def _convert_return_value(self, ret):
        """ convert return value to flask response value """
        status_code = 200
        if isinstance(ret, tuple) and len(ret) == 2 and isinstance(ret[1], int):
            status_code = ret[1]
            ret = ret[0]

        if isinstance(ret, str):
            # return string
            return ret, status_code
        elif isinstance(ret, Exception):
            name = ret.__class__.__name__
            ret = {'return_value': -1, 'exception': name + ': ' + str(ret)}
            return flask.jsonify(ret), 500
        elif hasattr(ret, 'html') and inspect.ismethod(getattr(ret, 'html')):
            # for Element, convert to html
            return ret.html(), status_code
        elif isinstance(ret, dict):
            return flask.jsonify(ret), 200
        elif isinstance(ret, list):
            return flask.jsonify(ret), 200
        elif isinstance(ret, int) or isinstance(ret, float):
            ret = {'return_value': ret}
            return flask.jsonify(ret), 200
        else:
            return str(ret), status_code

    def _decorator(self, func, defaults):
        """ decorate the func """
        def inner(*args):
            print('request.args', request.args)

            kwargs = OrderedDict()  # kwargs to call func
            for name in defaults:
                # get default value
                val = defaults[name]
                # if arg name exists in request
                # replace value with request
                if name in request.args:
                    val = request.args[name]
                kwargs[name] = val
            # call func with kwargs
            try:
                ret = func(**kwargs)
            except Exception as e:
                ret = e
            # analysis return value
            return self._convert_return_value(ret)

        return inner

    def _find_static_file(self, path):
        if self._static_folder and ServerUtil.is_valid_path(path):
            filename = self._static_folder + path
            if os.path.exists(filename):
                return filename
        return ''

    def _on_404_not_found(self, e):
        """ process when 404 not found"""
        file_ext = ServerUtil.file_ext(request.path)
        if file_ext not in HttpServer.FileExts:
            return e

        filename = self._find_static_file(request.path)
        if filename:
            # url = self._static_url_path + request.path
            # return flask.redirect(url)
            return flask.send_file(filename)
        else:
            return e

    def add_url(self, url_rule, view_func, endpoint=None):
        """
        add url rule

        :param url_rule: url rule, such as "/bin"
        :param view_func: function or method or object
        :param endpoint: (optinal) endpoint name str
        :return: self
        """
        self._confirm_app()

        if inspect.ismethod(view_func) or inspect.isfunction(view_func):
            # add url_rule from function
            # auto detect endpoint name
            if endpoint is None:
                if view_func:
                    url = url_rule if url_rule.endswith('/') else url_rule + '/'
                    endpoint = url + view_func.__name__

            if url_rule == '/':
                self._has_root = True

            params = self._get_func_params(view_func)
            func = self._decorator(view_func, params)
            self.app.add_url_rule(url_rule, endpoint, func, methods=['POST', 'GET'])
            # if params:
            #     func = self._decorator(view_func, params)
            #     self.app.add_url_rule(url_rule, endpoint, func)
            # else:
            #     self.app.add_url_rule(url_rule, endpoint, view_func)

        elif view_func is not None:
            # add url_rule from object
            obj = view_func

            if not url_rule.endswith('/'):
                url_rule += '/'

            for name, method in inspect.getmembers(obj):
                if not name.startswith('_'):
                    if inspect.ismethod(method):
                        if name == 'index':
                            url = url_rule
                        else:
                            url = url_rule + name
                        self.add_url(url, method)

        return self

    def _index_page(self):
        """ default home page of the server """
        return HttpServer.WelcomeHtml

    def static_folder(self, path, url=None):
        """ set static folder """
        if not (path.startswith("/") or path.find(':') == 1):
            path = os.path.join(os.getcwd(), path)

        if not os.path.exists(path):
            raise ValueError("Path %s not exists" % path)

        self._static_folder = path
        self._static_url_path = url
        return self

    def _prepare_static_folder(self):
        # static folder
        static_folder = self._static_folder
        if not static_folder:
            path = os.path.join(os.getcwd(), "www")
            if os.path.exists(path):
                static_folder = path
                self._static_folder = path

        if os.path.exists(static_folder):
            self.app.static_folder = static_folder
            if not self._static_url_path:
                self._static_url_path = 'static'
            self.app.static_url_path = self._static_url_path

    def run(self, threaded=False):
        """
        run the server

        :param threaded: (optional) whether run in thread
        :return: self
        """
        self._confirm_app()
        self._prepare_static_folder()

        # register 404 handler
        self.app.register_error_handler(404, self._on_404_not_found)

        # add default page
        if not self._has_root:
            self.add_url('/', self._index_page)

        if threaded:
            self._server = ServerThread(self.app, self.port, self.ip)
            self._server.start()
        else:
            self.app.run(host=self.ip, port=self.port, debug=self.debug)

        return self

    def stop(self):
        """ stop the server (available when run in thread) """
        if self._server:
            self._server.shutdown()
            self._server = None
            self.app = None
        return self


class ServerUtil:
    @staticmethod
    def file_ext(path):
        if isinstance(path, str):
            offset = path.find("?")
            path = path[:offset] if offset >= 0 else path
            offset = path.rfind('.')
            if offset:
                return path[offset:].lower()
        return ''

    @staticmethod
    def is_valid_path(path):
        if not isinstance(path, str):
            return False

        if path.find('..') >= 0:
            return False

        if re.search('[?$#]', path):
            return False

        return True


class ServerLogFilter(logging.Filter):
    def __init__(self, name='', log_file=None):
        super().__init__(name=name)
        self.log_file = log_file
        # self.logger = logging.Logger()

    def filter(self, rec):
        return False
        # filter all records except ERROR
        # return rec.levelno == logging.ERROR


class ServerThread(threading.Thread):
    # https://pyquestions.com/how-to-stop-flask-application-without-using-ctrl-c
    def __init__(self, flask_app, port, host='0.0.0.0', log_file=None):
        """  A server thread """
        super().__init__()
        self.log_file = log_file
        self.server = make_server(host, port, flask_app, threaded=True)
        self.change_logger()
        self.ctx = flask_app.app_context()
        self.ctx.push()

    def change_logger(self):
        logger = logging.getLogger("werkzeug")
        if logger:
            logger.addFilter(ServerLogFilter())
        # print(logger)
        # for handler in logger.handlers:
        #     print(handler)

        # # Create handlers
        # c_handler = logging.StreamHandler()
        # f_handler = logging.FileHandler('file.log')
        # c_handler.setLevel(logging.WARNING)
        # f_handler.setLevel(logging.INFO)
        #
        # # Create formatters and add it to handlers
        # c_format = logging.Formatter('%(name)s - %(levelname)s - %(message)s')
        # f_format = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')
        # c_handler.setFormatter(c_format)
        # f_handler.setFormatter(f_format)
        #
        # # Add handlers to the logger
        # logger.addHandler(c_handler)
        # logger.addHandler(f_handler)

    def run(self):
        self.server.serve_forever()

    def shutdown(self):
        self.server.shutdown()
